"""Parse tabular result files from common tools"""
import pandas as pd


def read_encyclopedia(proteins_txt: str) -> pd.DataFrame:
    """Read results from EncyclopeDIA.

    Parameters
    ----------
    proteins_txt : str
        The EncyclopeDIA protein output.

    Returns
    -------
    pandas.DataFrame
        The EncyclopeDIA results in a format for gopher.
    """
    proteins = pd.read_table(proteins_txt)
    accessions = proteins["Protein"].str.extract(r"\|(.+?)\|", expand=False)

    proteins = proteins.set_index(accessions)
    return proteins.drop(
        columns=["Protein", "NumPeptides", "PeptideSequences"]
    )


def read_metamorpheus(proteins_txt: str) -> pd.DataFrame:
    """Read results from Metamorpheus.

    Parameters
    ----------
    proteins_txt : str
        The Metamorpheus protein output file.

    Returns
    -------
    pandas.DataFrame
    """
    proteins = pd.read_table(proteins_txt, low_memory=False)
    accessions = proteins["Protein"].str.extract(
        r"^(.*?)(\||$)", expand=False
    )[0]
    accessions.name = "Protein"
    int_cols = [c for c in proteins.columns if c.startswith("Intensity")]
    proteins = proteins.set_index(accessions)
    proteins = (
        proteins.loc[
            proteins["Protein Decoy/Contaminant/Target"] == "T", int_cols
        ]
        .apply(pd.to_numeric)
        .fillna(0)
    )
    return proteins



"""Test that the enrichment functions are working correctly"""
import random

import numpy as np
import pandas as pd
from scipy import stats

import gopher
from gopher import annotations, enrichment


def test_entire_enrichment_analysis(generate_proteins):
    """Check that the test enrichment function returns a dataframe.
    This test will run the full GO enrichment on a dataset."""
    df = generate_proteins
    df.set_index("Protein", inplace=True)
    result = enrichment.test_enrichment(df)
    assert isinstance(result, pd.DataFrame)


def test_subset_enrichment_analysis(generate_proteins):
    """Check that the test enrichment function returns a dataframe.
    This test will run the GO enrichment on a subset of terms."""
    df = generate_proteins
    df.set_index("Protein", inplace=True)
    terms = [
        "nucleus",
        "nucleoplasm",
        "euchromatin",
        "heterochromatin",
        "protein-DNA complex",
        "lysosome",
        "cytoplasm",
    ]
    result = enrichment.test_enrichment(df, go_subset=terms)
    terms_found = result["GO Name"].unique()
    assert all(t in terms for t in terms_found)


def test_custom_annotations(generate_proteins):
    """Test that a custom annotations file gives a different output than the default and contains the new GO term,"""
    prot = generate_proteins
    custom = annotations.generate_annotations(
        proteins=prot["Protein"][: len(prot) // 2], aspect="C", go_name="test"
    )
    full, _ = annotations.load_annotations(species="human")
    annot = pd.concat([custom, full])
    prot.set_index("Protein", inplace=True)
    terms = [
        "nucleus",
        "nucleoplasm",
        "euchromatin",
        "heterochromatin",
        "protein-DNA complex",
        "lysosome",
        "cytoplasm",
        "test",
    ]
    custom_result = enrichment.test_enrichment(
        prot, annotations=annot, go_subset=terms, aggregate_terms=False
    )
    normal_result = enrichment.test_enrichment(
        prot, go_subset=terms, aggregate_terms=False
    )
    assert not custom_result.equals(normal_result)
    terms_found = custom_result["GO Name"].unique()
    assert all(t in terms for t in terms_found)
    assert "test" in terms_found


def test_custom_mapping_annotations(
    generate_fake_proteins, generate_annotations, generate_mapping
):
    """Test the custom mapping and custom annotations on fake data."""
    subset = ["b", "i", "z"]
    result = enrichment.test_enrichment(
        generate_fake_proteins,
        annotations=generate_annotations,
        mapping=generate_mapping,
        go_subset=subset,
    )
    assert result["GO ID"].values.tolist() == subset


def test_mannwhitneyu_small(generate_arrays):
    """Test the Mann-Whitney U numba function returns the same values
    as the scipy function on a small dataset for a two-sided test"""
    list1, list2 = generate_arrays
    res_scipy = stats.mannwhitneyu(list1, list2)
    res_numba = gopher.stats.mannwhitneyu(list1, list2)
    sp = np.round(res_scipy[1], 5)
    num = np.round(res_numba[1], 5)
    np.testing.assert_allclose(sp, num)


def test_mannwhitneyu_greater(generate_arrays):
    """Test the Mann-Whitney U numba function returns the same values
    as the scipy function on a small dataset for a one-sided, greater test"""
    list1, list2 = generate_arrays
    res_scipy = stats.mannwhitneyu(list1, list2, alternative="greater")
    res_numba = gopher.stats.mannwhitneyu(list1, list2, alternative="greater")
    sp = np.round(res_scipy[1], 5)
    num = np.round(res_numba[1], 5)
    np.testing.assert_allclose(sp, num)


def test_mannwhitneyu_less(generate_arrays):
    """Test the Mann-Whitney U numba function returns the same values
    as the scipy function on a small dataset for a one-sided, lesser test"""
    list1, list2 = generate_arrays
    res_scipy = stats.mannwhitneyu(list1, list2, alternative="less")
    res_numba = gopher.stats.mannwhitneyu(list1, list2, alternative="less")
    sp = np.round(res_scipy[1], 5)
    num = np.round(res_numba[1], 5)
    np.testing.assert_allclose(sp, num)


def test_mannwhitneyu_big():
    """Test the Mann-Whitney U numba function returns the same values
    as the scipy function on a larger dataset for a two-sided test"""
    list1 = []
    list2 = []
    random.seed(10)
    for i in range(0, 10):
        temp1 = []
        temp2 = []
        for i in range(0, 100):
            n = random.randint(1, 100)
            temp1.append(n)
            n = random.randint(50, 150)
            temp2.append(n)
        list1.append(temp1)
        list2.append(temp2)
    res_scipy = stats.mannwhitneyu(list1, list2)
    res_numba = gopher.stats.mannwhitneyu(list1, list2)
    sp = np.round(res_scipy[1], 10)
    num = np.round(res_numba[1], 10)
    np.testing.assert_allclose(sp, num)



"""Numba Mann-Whitney U test"""
import numba as nb
import numpy as np
from scipy import stats


@nb.njit(parallel=True)
def tiecorrect(rankvals: np.ndarray) -> np.ndarray:
    """Numba version of scipy.stats.tiecorrect

    Parameters
    ----------
    rankvals : np.ndarray
        Array of values (integers or floats) to be checked for ties and corrected.

    Returns
    -------
    np.ndarray
        Array of tie-corrected values.
    """
    tc = np.ones(rankvals.shape[1], dtype=np.float64)
    for j in nb.prange(rankvals.shape[1]):
        arr = np.sort(np.ravel(rankvals[:, j]))
        is_not_tie = np.concatenate(
            (np.array([True]), arr[1:] != arr[:-1], np.array([True]))
        )
        idx = np.nonzero(is_not_tie)[0]
        cnt = np.diff(idx).astype(np.float64)

        size = np.float64(arr.size)
        if size >= 2:
            tc[j] = 1.0 - (cnt**3 - cnt).sum() / (size**3 - size)

    return tc


@nb.njit(parallel=True)
def rankdata(data: np.ndarray) -> np.ndarray:
    """Numba version of scipy.stats.rankdata

    Parameters
    ----------
    data : np.ndarray
        Array of values (integers or floats) to be ranked.

    Returns
    -------
    np.ndarray
        Array of ranked values.
    """
    ranked = np.empty(data.shape, dtype=np.float64)
    for j in nb.prange(data.shape[1]):
        arr = np.ravel(data[:, j])
        sorter = np.argsort(arr)

        arr = arr[sorter]
        obs = np.concatenate((np.array([True]), arr[1:] != arr[:-1]))

        dense = np.empty(obs.size, dtype=np.int64)
        dense[sorter] = obs.cumsum()

        # cumulative counts of each unique value
        count = np.concatenate((np.nonzero(obs)[0], np.array([len(obs)])))
        ranked[:, j] = 0.5 * (count[dense] + count[dense - 1] + 1)

    return ranked


def mannwhitneyu(
    x: np.ndarray, y: np.ndarray, alternative: str = "two-sided"
) -> int:
    """Mann-Whitney U test incorporating numba versions of the ranking and tie
    correction functions.

    Parameters
    ----------
    x : np.ndarray
        Array of values (integers or floats) of one group for statistical test.
    y : np.ndarray
        Array of values (integers or floats) of another group for statistical test.
    alternative: str {"greater", "less", "two-sided"}
        Type of test that should be run.

    Returns
    -------
    Integer
        U-statistic from Mann-Whitney U test
    Integer
        p-value from Mann-Whitney U test
    """

    x = np.asarray(x)
    y = np.asarray(y)
    assert x.shape[1] == y.shape[1]

    n1 = x.shape[0]
    n2 = y.shape[0]

    # rank the data for the test
    ranked = rankdata(np.concatenate((x, y)))
    rankx = ranked[0:n1, :]  # get the x-ranks
    r = np.sum(rankx, axis=0)
    u1 = n1 * n2 + (n1 * (n1 + 1)) / 2.0 - r  # calc U for x
    u2 = n1 * n2 - u1  # remainder is U for y

    # check for ties in the rankings
    T = tiecorrect(ranked)

    # if *everything* is identical we'll raise an error, not otherwise
    if np.all(T == 0):
        raise ValueError("All numbers are identical")

    # get mean and standard deviation
    sd = np.sqrt(T * n1 * n2 * (n1 + n2 + 1) / 12.0)
    meanrank = n1 * n2 / 2.0 + 0.5

    # do a one-sided or two-sided test depending on user specification
    if alternative == "greater":
        U, f = u2, 1
    elif alternative == "less":
        U, f = u1, 1
    else:
        U, f = np.maximum(u1, u2), 2

    # calculate z and use it to find the p-values
    with np.errstate(divide="ignore", invalid="ignore"):
        z = (U - meanrank) / sd

    p = stats.norm.sf(z)
    p *= f
    p = np.clip(p, 0, 1)

    return U, p



"""See the README for detailed documentation and examples."""
try:
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version(__name__)
    except PackageNotFoundError:
        pass

except ImportError:
    from pkg_resources import DistributionNotFound, get_distribution

    try:
        __version__ = get_distribution(__name__).version
    except DistributionNotFound:
        pass

from .annotations import generate_annotations, load_annotations
from .config import get_data_dir, set_data_dir
from .display_data import (
    get_annotations,
    get_rankings,
    in_term,
    map_proteins,
    roc,
)
from .enrichment import test_enrichment
from .normalize import normalize_values
from .parsers import read_encyclopedia, read_metamorpheus
